<!-- 
	Generates the metadata for a PDF document. 
-->
<xsl:stylesheet 
    xmlns:xsl="http://www.w3.org/1999/XSL/Transform" 
    xmlns:xs="http://www.w3.org/2001/XMLSchema" 
    xmlns:fo="http://www.w3.org/1999/XSL/Format" 
    xmlns:oxy="http://www.oxygenxml.com/css2fo" 
    xmlns:css="http://www.w3.org/1998/CSS"
    xmlns:dc="http://purl.org/dc/elements/1.1/"
    xmlns:xmp="http://ns.adobe.com/xap/1.0/"
    xmlns:x="adobe:ns:meta/"
    xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
    xmlns:adobe-pdf="http://ns.adobe.com/pdf/1.3/"
    xmlns:pdf="http://xmlgraphics.apache.org/fop/extensions/pdf"
    xmlns:nd="http://www.oxygenxml.com/css2fo/named-destinations"
    
    exclude-result-prefixes="xs oxy dc xmp x adobe-pdf pdf rdf nd" version="2.0"
    xpath-default-namespace="http://www.w3.org/1998/CSS">
    
    <!--
        Collects the relevant properties that are set across the document, prefering the first 
        ocurrence of each one and sets them in a XMP structure.

		Context: The root element.
      -->    
    <xsl:template name="generate-metadata">
        
        <!-- Metadata that appears in the Acrobat Reader "Document Properties"/"Description" page -->
        <xsl:variable name="title" select="(//css:oxy-pdf-meta-title)[1]"/>
        <xsl:variable name="author" select="(//css:oxy-pdf-meta-author)[1]"/>
        <xsl:variable name="description" select="(//css:oxy-pdf-meta-description)[1]"/>
        <xsl:variable name="keywords" select="(//css:oxy-pdf-meta-keywords)[1]"/>
        
        <xsl:variable name="keyword-set" select="//css:oxy-pdf-meta-keyword"/>
        
        <fo:declarations>
            <x:xmpmeta>
                <rdf:RDF>
                    <rdf:Description>

                        <xsl:if test="$title">
                            <dc:title>
                                <xsl:value-of select="oxy:get-all-static-as-text($title)"/>
                            </dc:title>
                        </xsl:if>

                        <xsl:if test="$author">
                            <dc:creator>
                                <xsl:value-of select="oxy:get-all-static-as-text($author)"/>
                            </dc:creator>
                        </xsl:if>

                        <xsl:if test="$description">
                            <dc:description>
                                <xsl:value-of select="oxy:get-all-static-as-text($description)"/>
                            </dc:description>
                        </xsl:if>

						<xsl:choose>
						    <xsl:when test="$keywords">
						        <adobe-pdf:Keywords>
						            <xsl:value-of select="oxy:get-all-static-as-text($keywords)"/>
						        </adobe-pdf:Keywords>
						    </xsl:when>
						    <xsl:when test="$keyword-set">
						        <adobe-pdf:Keywords>
					                <xsl:for-each select="$keyword-set">
					                    <xsl:value-of select="oxy:get-all-static-as-text(.)"/>
					                    <xsl:if test="position() != last()">, </xsl:if>
					                </xsl:for-each>
						        </adobe-pdf:Keywords>
						    </xsl:when>
						</xsl:choose>
                        
                        
                        <xmp:CreatorTool>
                        
                          <xsl:variable name="creator">
                            <xsl:value-of select="$application-name"/>
                            
                            <xsl:if test="$application-version">
                              Version <xsl:value-of select="$application-version"/>                            
                            </xsl:if>
                            
                            <xsl:if test="$application-build-timestamp">
                              Build <xsl:value-of select="$application-build-timestamp"/>                            
                            </xsl:if>
                          </xsl:variable>
                          
                          <xsl:value-of select="normalize-space($creator)"/>
                          
                        </xmp:CreatorTool>
                    </rdf:Description>
                </rdf:RDF>
            </x:xmpmeta>

            <!-- Metadata that appears in the Acrobat Reader "Document Properties"/"Custom" page -->
            <xsl:variable name="custom-meta-names" select="distinct-values(//oxy-pdf-meta-custom/@name)"/>
            <xsl:if test="exists($custom-meta-names)">
                <pdf:info>
                    <xsl:variable name="doc" select="/"/>
                    <xsl:for-each select="$custom-meta-names">
                        <xsl:variable name="value" select="oxy:get-all-static-as-text(($doc//oxy-pdf-meta-custom[@name = current()])[1])"/>
                        <pdf:name key="{.}"><xsl:value-of select="$value"/></pdf:name>
                    </xsl:for-each>
                </pdf:info>
            </xsl:if>

            <xsl:if test="$title">
                <pdf:catalog>
			        <!-- <pdf:entry key="PageMode" type="name">FullScreen</pdf:entry> -->
			        <!-- <pdf:entry key="PageLayout" type="name">SinglePage</pdf:entry> -->
                    <pdf:dictionary key="ViewerPreferences">
                        <!-- <pdf:entry key="HideToolbar" type="boolean">true</pdf:entry> -->
                        <!-- <pdf:entry key="HideWindowUI" type="boolean">true</pdf:entry> -->
                        <pdf:boolean key="DisplayDocTitle">true</pdf:boolean>
                        <!-- <pdf:entry key="NonFullScreenPageMode" type="name">UseThumbs</pdf:entry> -->
                    </pdf:dictionary>
                </pdf:catalog>
            </xsl:if>
        </fo:declarations>

       <xsl:call-template name="generate-named-destination-list"/>
        
    </xsl:template>
    
    <!-- Do not propagate the meta attributes to the FO, are useless there. -->
    <xsl:template match="css:oxy-pdf-meta-title" priority="2"/>
    <xsl:template match="css:oxy-pdf-meta-author" priority="2"/>
    <xsl:template match="css:oxy-pdf-meta-description" priority="2"/>
    <xsl:template match="css:oxy-pdf-meta-keywords" priority="2"/>
    <xsl:template match="css:oxy-pdf-meta-keyword" priority="2"/>
    <xsl:template match="css:oxy-pdf-meta-custom" priority="2"/>
    
</xsl:stylesheet>